/*****************************************************************************\
|*
|*  Version:    1.5
|*
|*  Description:
|*
|*      FLASH functions to be ran from RAM (N.B. do NOT access ROM
|*      in these functions)
|*
|*  Copyright 2003 Altium BV
|*
 *****************************************************************************/

#include "target.h"
#include "uart.h"
#include "command.h"
#include "flash.h"


#define pragma_string(x) _Pragma( #x )
#define pragma_preprocess(arg) pragma_string(arg)
pragma_preprocess(renamesect CO=flash init __at(FLASH_AREA_ADDR))
/*
 * This pragma causes that all functions below are put in a segment
 * flash_CO with the INIT segment attribute. This attribute ensures
 * that the actual functions are located in RAM (from where they
 * should run), but at the same time a copy is allocated in ROM.
 * Normally the RAM area is initialized during startup by the __init
 * routine, but it can be initialized at anytime using the following
 * locator labels:
 *   _lc_b_flash_CO     start address of the flash_CO segment in RAM
 *   _lc_e_flash_CO     end   address of the flash_CO segment in RAM
 *   _lc_cb_flash_CO    start address of the copy flash_CO segment in ROM
 *   _lc_ce_flash_CO    start address of the copy flash_CO segment in ROM
 * Assuming the RAM area is in the first 64k the special SMOVFW/SMOVFB
 * instruction can be used to perform the copy:
 *    __smovfw(_lc_cb_flash_CO,                                 // source (__far)
 *             _lc_b_flash_CO,                                  // dest (__near)
 *             (_lc_e_flash_CO - _lc_b_flash_CO + 1) >> 1 );    // size
 *
 * See the 'm16c_write_flash()' and 'm16c_erase_flash()' functions
 * in the 'm16c.c' module for an example.
 *
 */

/**********************************************************************\
|*
|*  FUNCTION: ram_erase_flash
|*
|*  DESCRIPTION: erase flash memory (except for blocks containing
|*               ROM monitor offcourse).
|*
 */
void ram_erase_flash(void)
{
    ta_slow_down_clock();   
    ta_erase_flash();
    ta_speed_up_clock();   
}

/**********************************************************************\
|*
|*  FUNCTION: ram_uart_putchar
|*
|*  PARAMETERS: byte, byte for output
|*
|*  DESCRIPTION: write a byte to the UART
|*
 */
void ram_uart_putchar(unsigned char byte)
{
    while( !UxC1_TI )
    {
        /* NOP */ ;
    }
    UxTB = byte;
}

/**********************************************************************\
|*
|*  FUNCTION: ram_write_flash
|*
|*  PARAMETERS: address, start address where data is to be flashed
|*              buf, buffer with 256 bytes
|*
|*  DESCRIPTION: write flash memory.
|*               if needed, the target specific functions calculate
|*               the actual pages to use. Input data does not need to be
|*               aligned. Where no data is defined, value 0xFF will be used
|*               to avoid damaging data already writen...
|*
 */
void ram_write_flash(unsigned long address, __near unsigned char *buf )
{
    ta_slow_down_clock();   
    ta_write_flash(address,buf);
    ta_speed_up_clock();   
}

#ifdef DEBUG
/**********************************************************************\
|*
|*  Function:      ram_uart_putnibble
|*
|*  Parameters:    byte = byte containing nibble to be sent
|*
|*  Description:   send a nibble in hex format (0-9A-F) through the serial port
|*
 */
void
ram_uart_putnibble( unsigned char byte )
{
    byte &= 0xf;
    while( !UxC1_TI )
    {
        /* NOP */ ;
    }
    if ( byte < 10 )
    {
        UxTB = byte + '0';
    }
    else
    {
        UxTB = byte + 'A'-10;
    }
}
#endif
