/*****************************************************************************
|*
|*  Version:    1.1
|*
|*  Description:
|*
|*      Target specific functions/settings for the Glyn M30260 ROM monitor
|*
|*  Copyright 2002-2005 Altium BV
|*
 *  M30260F8A memory MAP
 *
 *  64k rom, 2k ram
 *             _________
 *     0x0000 |         |
 *            |         |
 *            |   SFR   |
 *            |         |
 *            |_________|  _
 *     0x0400 |         |   \
 *            |    f    |   |
 *            |    r    |   |
 *            |    e    |   |
 *            |    e    |   +-- free to use for user application
 *            |         |   |
 *            |         |   |
 *            |         |   |
 *            |_________|  _/
 *     0x0800 |         |   \
 *            |RAM funcs|   |
 *            |_________|   +-- may be used by user app. but will be
 *     0x0920 |         |   |   overwritten during download and flashing
 *            |dwnh buf.|   |
 *            |_________|  _/
 *     0x0B20 |   ROM   |
 *            | monitor |  <- Must always be reserved in user application
 *            |  vars.  |
 *     0x0BFF |_________|
 *
 *****************************************************************************/

#ifndef _GLYN_M30260_H
#define _GLYN_M30260_H

/*
 * defines, typedefs
 */
#define FOSC                    20000000L
#define VERSION                 "V3.0"
#define UART                    1
#define BAUDRATE                57600
#define BRG                     ((FOSC/(BAUDRATE*16))-1)
#define FCLOCK                  0
#define HANDSHAKE_SOFTWARE
#define SERIAL_VECTOR_STR       "0FE000"
#define ROM_MONITOR_START       0xFE000
#define USER_AREA_ADDR          0x400        /* start address of the area to be used by the user application */
#define FLASH_AREA_ADDR         0x800        /* start address of the area used by flash functions */
#define DOWNLOAD_BUF_ADDR       0x920        /* start address of the 512 byte buffer used during download of bytes to flash */
#define USER_AREA_SIZE          (FLASH_AREA_ADDR-USER_AREA_ADDR)

#include "uart.h"       /* in this file SERIAL_INT_STR will be set */
#include "command.h"
#include "flash.h"

/***************************************************************
 *
 * Configuration string, answered at CMD_CONF
 *
 * The following parameters can be passed:
 *   CPU  = CPU type
 *   PM   = Processor mode the ROM monitor is running in, can be one of:
 *           SC = Single-chip mode
 *           ME = Memory expansion mode
 *           MP = Microprocessor mode
 *   OP   = EX => expand internal RAM and ROM areas
 *   SN   = serial transmit interrupt number used by the ROM monitor
 *          it is assumed that the serial receive interrupt is one higher.
 *   SV   = serial vector address, which is the address of the actual
 *          serial interrupt handler (same for receive and transmit).
 *   In any user application running under the ROM monitor it is
 *   essential that the serial interrupt vector keeps pointing to
 *   interrupt handler routine otherwise communication with the
 *   ROM monitor will hang. Thus if the user application uses its own
 *   vector table the 'SN'th and 'SN+1'th entry in this table should contain value 'SV'!
 *   RAM    = <start>,<size>  RAM area(s) present on the target board
 *   ROM    = <start>,<size>  ROM area(s) present on the target board
 *   FLASH  = <start>,<size>  Flash area(s) present on the target board
 *   RE     = <start>,<size>  Area(s) used by the ROM monitor, these need
 *                            to be reserved by the User application
 *   BP     = <integer>       Number of breakpoints supported by monitor.
 *                            When omitted 2 is assumed.
 */
#define CONF_STR      VERSION " TASKING M16C ROM monitor\n\
TG=Glyn M16C/26A\n\
CPU=M30260\n\
PM=SC\n\
SN=" SERIAL_INT_STR    "\n\
SV=" SERIAL_VECTOR_STR "\n\
RE=0B20,E0\n\
RE=FE000,2000\n\
TG=GLYN M30260"

/*
 * global functions, inline functions
 */

/**********************************************************************
|*
|*  Function: ta_reset
|*
|*  Parameters: void
|*
|*  Description: generate MCU reset. 
|*
|*  RETURNS:    void (MCU is reset...)
|*              
 */
inline void ta_reset(void)
{
    PRC1 = 1; /* enable writing processor mode register */
    PM03  = 1; /* set SW reset bit */
    PRC1 = 0; /* done... */
}

/**********************************************************************
|*
|*  Function: ta_init
|*
|*  Parameters: void
|*
|*  Description: additional init code. 
|*
|*  RETURNS:    void
|*              
 */
inline void ta_init(void)
{
}

/**********************************************************************
|*
|*  Function:      ta_show_conf
|*
|*  Description:
|*    Send the monitor configuration string.
|*
 */
inline void ta_show_conf(void)
{
    uart_puts( CONF_STR );
}

/**********************************************************************
|*
|*  Function: ta_dummy_user_app
|*
|*  Description: 
|*     Function used as dummy user application when a 'GO' command is
|*     issued before an application is downloaded and before the
|*     appropriate Program Counter is set.
|*              
 */
inline void ta_dummy_user_app( void )
{
    while ( 1 )
    {
        uart_putchar( '+' );
    }
}

/**********************************************************************\
|*
|*  FUNCTION: ta_slow_down_clock
|*
|*  DESCRIPTION: slow down the system clock to 1/8
|*
 */
inline void ta_slow_down_clock( void )
{
    PRC0 = 1;
    CM06 = 1;      /* set clk/8 */
    PRC0 = 0;
}

/**********************************************************************\
|*
|*  FUNCTION: ta_speed_up_clock
|*
|*  DESCRIPTION: speed up the system clock
|*
 */
inline void ta_speed_up_clock( void )
{
    /* TODO user application may use different clock,
     * actually we should restore that clock rate
     */
    PRC0 = 1;
    CM06 = 0;       /* set clk/1 */
    PRC0 = 0;
}

/**********************************************************************
|*
|*  Function: ta_erase_block
|*
|*  Parameters: x, highest even address in block to erase
|*
|*  Description: erase a specific block of flash memory
|*              
 */
inline void ta_erase_block(unsigned long x)
{ 
    while ( !FMR00 ) 
    { 
    } 
    *(unsigned int *)(x) = 0x0020;  
    *(unsigned int *)(x) = 0x00D0;  
}    

/**********************************************************************
|*
|*  Function: ta_erase_flash
|*
|*  Description: erase flash memory 
|*               the blocks array identifies which blocks are to be
|*               erased (not the blocks where the rom monitor is !!)
|*
 */
inline void ta_erase_flash(void)
{
    /* enable Rewrite Mode */
    FMR01 = 0;
    FMR01 = 1;
        
    /* disable lock bits */
    FMR02 = 0;
    FMR02 = 1;

    FMR16 = 0;
    FMR16 = 1;

    ta_erase_block( 0xf7ffe );
    ta_erase_block( 0xfbffe );
    ta_erase_block( 0xfdffe );

    while ( !FMR00 )
    {
    }   /* wait until erase is ready */

    /* enable lock bits */
    FMR02 = 0;

    FMR16=0;

    /* reset flash memory */
    FMSTP = 1;
    FMSTP = 0;
        
    /* disable Rewrite Mode */
    FMR01 = 0;
}

/**********************************************************************
|*
|*  Function: ta_write_flash
|*
|*  Description: write flash memory 
|*               memory is written in pages of 256 byte
|*
|*                :             :             :             :             :
|*      in flash: +---------------------------+---------------------------+
|*                | first page with 256 bytes | second 256 byte page      |
|*                +---------------------------+---------------------------+
|*
|*
 */
inline void ta_write_flash(unsigned long address, __near unsigned char * buf)
{
    unsigned int index;

    /* enable Rewrite Mode */
    FMR01 = 0;
    FMR01 = 1;

    FMR02 = 0;
    FMR02 = 1;

    FMR16 = 0;
    FMR16 = 1;


    for (index=0; index<DOWNLOADBUFSIZE/2; index++)
    {
        /* For HND type flashes every write should be preceded with
         * a write of the constant value 0x0040.
         */
        ((__far unsigned int *)address)[index] = 0x0040;

        ((__far unsigned int *)address)[index] = ((__near unsigned int *)buf)[index];

        while ( !FMR00 )
        {
            /* wait until action done */ ;
        }

        /* reset flash memory */
        FMSTP = 1;
        FMSTP = 0;
    }
    FMR02 = 0;

    FMR16=0;

    /* reset flash memory */
    FMSTP = 1;
    FMSTP = 0;
    /* disable Rewrite Mode */
    FMR01 = 0;
}

#endif /* _GLYN_M30260_H */
