/*****************************************************************************\
|*
|*  Version:    1.7
|*
|*  Description:
|*
|*      TASKING M16C ROM Monitor: M16C specific functions
|*
|*  Copyright 2003 Altium BV
|*
 *****************************************************************************/

#include "target.h"
#include "uart.h"
#include "command.h"
#include "m16c.h"
#include "flash.h"

#define REIT    0xfb

/*
 * local function prototypes
 */

unsigned char m16c_put_escaped_char(unsigned char c);
void m16c_stopped_at( void );

/*
 * variable declarations
 */
const unsigned char hex[]="0123456789ABCDEF";

/**********************************************************************\
|*
|*  Function:    m16c_get_FLG
|*
|*  Description: return the FLG contents
|*              
 */
unsigned int m16c_get_FLG( void )
{
    unsigned int flg;

    flg = isp_user[9] & 0xf0ff;

    return flg;
}

/**********************************************************************\
|*
|*  Function:    m16c_set_FLG
|*
|*  Description: set the FLG contents
|*              
 */
void m16c_set_FLG( unsigned int flg )
{
    isp_user[9] &= 0x0f00;
    isp_user[9] |= (flg & 0xf0ff);
}

/**********************************************************************\
|*
|*  Function:    m16c_get_PC
|*
|*  Description: return the program counter value
|*              
 */
unsigned long m16c_get_PC( void )
{
    unsigned int  userPCl;
    unsigned char userPCh;

    userPCl = isp_user[8];
    userPCh = (isp_user[9] >> 8) & 0xf;

    return (unsigned long)userPCl | (unsigned long)userPCh<<16;
}

/**********************************************************************\
|*
|*  Function:    m16c_set_PC
|*
|*  Description: set the program counter
|*              
 */
void m16c_set_PC( unsigned long pc )
{
    isp_user[8] = (unsigned int)pc;
    isp_user[9] &= 0xf0ff;
    isp_user[9] |= (unsigned int)((pc & 0xf0000L) >> 8);
}

/**********************************************************************\
|*
|*  Function: m16c_fill
|*
|*  Parameters: address, where to start filling
|*              length, number of items to write
|*              data_type, '0' for bytes, otherwise words
|*              data, uint value (in case of data_type '0' LSB is used)
|*
|*  Description: single step 'number' times
|*              
 */
void m16c_fill(unsigned long address, unsigned int length, unsigned char data_type, unsigned int data)
{
    if (data_type == '0')
    {
        for (unsigned int i = length; i > 0; i--)
        {
            *(__far unsigned char *)address = (unsigned char)data;
            ++address;
        }
    }
    else
    {
        for (unsigned int i = length; i > 0; i--)
        {
            *(__far unsigned int *)address = data;
            address += 2;
        }
    }
}

/**********************************************************************\
|*
|*  Function: m16c_put_escaped_char
|*
|*  Parameters: c, character to write to UART
|*
|*  Description: write character and escape when needed
|*
|*  Returns:    increment value for checksum
|*              
 */
unsigned char m16c_put_escaped_char(unsigned char c)
{
    unsigned char ret;

    ret = c;
    /* need to escape ? write escape + char+0x40 */
    if ((c == CTRL_ESC) || (c == CTRL_INIT) || (c == XOFF) || (c == XON))
    {
        uart_putchar( CTRL_ESC );
        c += 0x40;
        ret = ret + CTRL_ESC + 0x40;
    }
    uart_putchar( c );

    return ret;
}

/**********************************************************************\
|*
|*  Function: m16c_dump
|*
|*  Parameters: address, start address wher dumping starts
|*              length, numer of items to dump
|*              data_type, 0 (value not ascii char) for bytes, otherwise words
|*
|*  Description: dump 'length' elements from memory starting at 'address'
|*              
 */
void m16c_dump(unsigned int length, unsigned char data_type)
{
    unsigned int  i;
    unsigned char checksum = 0;

    for (i = (unsigned int) length; i > 0; i--)
    {
        checksum -= m16c_put_escaped_char( *(unsigned char *)(dump_address++) );
        if (data_type != 0) /* TODO check order of bytes in case of word dump */
        {
            checksum -= m16c_put_escaped_char( *(unsigned char *)(dump_address++) );
        }
    }
    /* output 2 nibbles checksum */
    uart_putchar( hex[(checksum & 0xF0) >> 4]);
    uart_putchar( hex[checksum & 0x0F]);
}

/**********************************************************************\
|*
|*  Function: m16c_step
|*
|*  Parameters: number, number of steps to make
|*
|*  Description: single step 'number' times
|*
|*  Returns:    ERR_NORMAL: all is fine
|*              ERR_IRREGULAR: extraordinary status
|*              
 */
unsigned char m16c_step(unsigned int * number)
{
    /* application goes... let the single step interrupt stop the app. */
    application_running = 1;

    /* A problem in the M16C core is that when single stepping over a REIT
     * instruction the user applications FLG setting will overwrite the Debug
     * flag, hence no single-step interrupt is generated after the REIT.
     * To workaround this check if we happen to step over a REIT, if so patch
     * the Debug flag in the user's FLG stored on stack.
     */
    if ( *(__far unsigned char *)m16c_get_PC() == REIT )
    {
        /* Software interrupts above 31 may have used the
         * USP instead of ISP, so check that first.
         */
        if ( isp_user[9] & 0x80 )
        {
            __near unsigned int *usp;
            __fset( __U );
            usp = (__near unsigned int *)__stc_sp();
            usp[1] |= 0x1 << __D;  /* set the Debug flag */
            __fclr( __U );
        }
        else
        {
            isp_user[11] |= 0x1 << __D;  /* set the Debug flag */
        }
    }
    /* need to set debug flag for single step mode and secret bit 0x8000 */
    isp_user[9] |= 0x8002;

    /* number of steps to take.. */
    step_count = *number;

    return ERR_NORMAL;
}

/**********************************************************************\
|*
|*  Function: m16c_write_flash
|*
|*  Parameters: buf, byte buffer to write to flash memory
|*
|*  Description: write data to flash memory
|*
 */
void m16c_write_flash(unsigned long address, __near unsigned char *buf)
{
    /* copy flash functions from rom to ram */
        __smovfw((__far int *)_lc_cb_flash_CO,
                 (__near int *)_lc_b_flash_CO,
                 (_lc_e_flash_CO - _lc_b_flash_CO + 1) >> 1);

    __fclr( __I );        /* disable interrupts */
    __nop();
    __nop();

    ram_write_flash(address, buf);

    __fset( __I );        /* enable interrupts */
    __nop();
    __nop();
}

/**********************************************************************\
|*
|*  Function: m16c_erase_flash
|*
|*  Description: erase flash memory
|*
 */
void m16c_erase_flash(void)
{
    /* copy flash functions from rom to ram */
    __smovfw((__far int *)_lc_cb_flash_CO,
             (__near int *)_lc_b_flash_CO,
             (_lc_e_flash_CO - _lc_b_flash_CO + 1) >> 1);

    __fclr( __I );        /* disable interrupts */

    ram_erase_flash();

    __fset( __I );        /* enable interrupts */
}

/**********************************************************************\
|*
|*  Function:     m16c_dump_reg
|*
|*  Parameters:   reg_number
|*
|*  Description:  get value of register 'reg_number'
|*
|*  Returns:      value of register
|*
 */
unsigned long m16c_dump_reg(unsigned char reg_number)
{
    if ( reg_number < 7 )
    {
        /* Front registers (bank flag = 0) */
        if ( isp_user[9] & (0x1 << __B) )
        {
            switch ( reg_number )
            {
            case 0:
                __fclr( __B );
                asmgetR0();
                __fset( __B );
                break;
            case 1:
                __fclr( __B );
                asmgetR1();
                __fset( __B );
                break;
            case 2:
                __fclr( __B );
                asmgetR2();
                __fset( __B );
                break;
            case 3:
                __fclr( __B );
                asmgetR3();
                __fset( __B );
                break;
            case 4:
                __fclr( __B );
                asmgetA0();
                __fset( __B );
                break;
            case 5:
                __fclr( __B );
                asmgetA1();
                __fset( __B );
                break;
            case 6:
                __fclr( __B );
                asmgetFB();
                __fset( __B );
                break;
            }
        }
        else if (reg_number < 6)
        {
            val4 = isp_user[reg_number];
        }
        else
        {
            val4 = isp_user[7];
        }
    }
    else if ( reg_number < 14 )
    {
        /* Back registers (bank flag = 1) */
        if ( !(isp_user[9] & (0x1 << __B)) )
        {
            switch ( reg_number )
            {
            case REG_R0:        /* 0 */
                __fset( __B );
                asmgetR0();
                __fclr( __B );
                break;
            case REG_R1:        /* 1 */
                __fset( __B );
                asmgetR1();
                __fclr( __B );
                break;
            case REG_R2:        /* 2 */
                __fset( __B );
                asmgetR2();
                __fclr( __B );
                break;
            case REG_R3:        /* 3 */
                __fset( __B );
                asmgetR3();
                __fclr( __B );
                break;
            case REG_A0:        /* 4 */
                __fset( __B );
                asmgetA0();
                __fclr( __B );
                break;
            case REG_A1:        /* 5 */
                __fset( __B );
                asmgetA1();
                __fclr( __B );
                break;
            case REG_FB:        /* 6 */
                __fset( __B );
                asmgetFB();
                __fclr( __B );
                break;
            }
        }
        else if (reg_number < 13)
        {
            val4 = isp_user[reg_number];
        }
        else
        {
            val4 = isp_user[7];
        }
    }
    else
    {
        switch ( reg_number )
        {
        case REG_USP:   /* 0xe */
            __fset( __U );
            val4 = __stc_sp();
            __fclr( __U );
            break;
        case REG_ISP:   /* 0xf */
            val4 = (unsigned int)isp_user + 20;
            break;
        case REG_FLG:   /* 0x10 */
            val4 = m16c_get_FLG();
            break;
        case REG_PC:    /* 0x11 */
            val56 = m16c_get_PC();
            break;
        case REG_SB:    /* 0x12 */
            val4 = isp_user[6];
            break;
        case REG_INTB:  /* 0x13 */
            val56 = ((unsigned long)__stc_intbh() << 16) | __stc_intbl();
            break;
        default:
            err_code[1] = SYNTAX_ERROR;
            return 0;
        }
    }
    /* only PC and INTB have 6 nibbles, other registers have 4 */
    if (( reg_number == 0x11 ) || (reg_number == 0x13 ))
    {
        uart_putchar(hex[ (val56 >> 20) & 0x0000000F ]);
        uart_putchar(hex[ (val56 >> 16) & 0x0000000F ]);
        val4 = (unsigned int)val56;
    }
    uart_putchar(hex[ (val4 >> 12) & 0x0000000F ]);
    uart_putchar(hex[ (val4 >>  8) & 0x0000000F ]);
    uart_putchar(hex[ (val4 >>  4) & 0x0000000F ]);
    uart_putchar(hex[ (val4 >>  0) & 0x0000000F ]);

    return 0;
}

/**********************************************************************\
|*
|*  Function: m16c_set_reg
|*
|*  Parameters: reg_number, register to set
|*              reg_value, value to set register to
|*
|*  Description: set value of register 'reg_number' to 'reg_value'
|*
 */
void m16c_set_reg( unsigned char reg_number, unsigned long reg_value)
{
    if ( reg_number < 6 )
    {
        isp_user[reg_number] = (unsigned int)reg_value;
    }
    else if ( reg_number == REG_FB /* 6 */)
    {
        isp_user[7] = (unsigned int)reg_value;
    }
    else if ( reg_number < 13 )
    {
        isp_user[reg_number-7] = (unsigned int)reg_value;
    }
    else
    {
        switch (reg_number)
        {
        case 0x0d: /* FB back */
            isp_user[7] = (unsigned int)reg_value;
            break;
        case 0x0e: /* USP */
            __fset( __U );
            __ldc_sp((unsigned int)reg_value);
            __fclr( __U );
            break;
        case 0x0f: /* ISP */
            __smovfw((__far int *)isp_user,(__near int *)reg_value, 10 );
            isp_user = (__near unsigned int *)reg_value;
            break;
        case 0x10: /* FLG */
            m16c_set_FLG((unsigned int)reg_value);
            break;
        case 0x11: /* PC */
            m16c_set_PC(reg_value);
            break;
        case 0x12: /* SB */
            isp_user[6] = (unsigned int)reg_value;
            break;
        case 0x13: /* INTB */
            __ldc_intbh((unsigned int)(reg_value >> 16));
            __ldc_intbl((unsigned int)reg_value);
            break;
        default:
            break;
        }
    }
}

/**********************************************************************\
|*
|*  Function: m16c_address_match
|*
|*  Description: address match interrupt routine
|*
 */
static void 
__frame(R0,R1,R2,R3,A0,A1,SB,FB) __interrupt_fixed(5)
m16c_address_match(void)
{
    application_running = 0;
    switch_to_monitor_mode();

    pc_pc = m16c_get_PC();      /* address of instruction after BP */

    pc_rmad0 = (unsigned long)RMAD0 & 0x000FFFFF;
    pc_rmad1 = (unsigned long)RMAD1 & 0x000FFFFF;
#ifdef HAS_4BREAKPOINTS
    pc_rmad2 = (unsigned long)RMAD2 & 0x000FFFFF;
    pc_rmad3 = (unsigned long)RMAD3 & 0x000FFFFF;
#endif
    /* check which breakpoint caused this interrupt */
    if ((AIER0) && ( ((pc_pc-1) == pc_rmad0) || ((pc_pc-2) == pc_rmad0)) )
    {
        pc_pc = pc_rmad0;
    }
    else if ( (AIER1) && ( ((pc_pc-1) == pc_rmad1) || ((pc_pc-2) == pc_rmad1)) )
    {
        pc_pc = pc_rmad1;
    }
#ifdef HAS_4BREAKPOINTS
    else if ((AIER20) && ( ((pc_pc-1) == pc_rmad2) || ((pc_pc-2) == pc_rmad2)) )
    {
        pc_pc = pc_rmad2;
    }
    else if ( (AIER21) && ( ((pc_pc-1) == pc_rmad3) || ((pc_pc-2) == pc_rmad3)) )
    {
        pc_pc = pc_rmad3;
    }
    AIER2 = 0; /* clear int. enable, othwerwise we would be stuck here forever */
#endif
    AIER = 0; /* clear int. enable, othwerwise we would be stuck here forever */
    

    m16c_set_PC(pc_pc); /* set correct PC to continue */

    m16c_stopped_at();
}

/**********************************************************************\
|*
|*  Function: m16c_single_step
|*
|*  Description: single step interrupt routine
|*
 */
static void 
__frame(R0,R1,R2,R3,A0,A1,SB,FB) __interrupt_fixed(4)
m16c_single_step(void)
{
    /* clear D-flag ? */
    step_count--;
    if (step_count==0)
    {
        application_running = 0;
        switch_to_monitor_mode();
        isp_user[9] &= ~(0x1 << __D);
        m16c_stopped_at();
    }
    else
    {
        isp_user[9] |= 0x8002;
    }
}

/**********************************************************************\
|*
|*  Function: m16c_break_instruction
|*
|*  Description: BRK instruction interrupt routine
|*
 */
static void 
__frame(R0,R1,R2,R3,A0,A1,SB,FB) __interrupt_fixed(6)
m16c_break_instruction(void)
{
    if ( user_mode == 1 )
    {
        /* the user application hit a breakpoint */
        switch_to_monitor_mode();

        /* PC points at BRK + 1 !!! */
        m16c_set_PC( m16c_get_PC()-1 );

        m16c_stopped_at();
    }
    else
    {
        AIER = 0x00;
#ifdef HAS_4BREAKPOINTS
        AIER2 = 0;
#endif
        if ( step_or_go == GO )
        {
            if  (!(break_points[0] & 0x80000000))
            {
                RMAD0L = (break_points[0] & 0x0ff);
                RMAD0M = (break_points[0] >> 8) & 0x0ff;
                RMAD0H = (break_points[0] >> 16) & 0x0f;
                AIER0 = 1;
            }
            if (!(break_points[1] & 0x80000000))
            {
                RMAD1L = (break_points[1] & 0x0ff);
                RMAD1M = (break_points[1] >> 8) & 0x0ff;
                RMAD1H = (break_points[1] >> 16) & 0x0f;
                AIER1 = 1;
            }
#ifdef HAS_4BREAKPOINTS
            if  (!(break_points[2] & 0x80000000))
            {
                RMAD2L = (break_points[2] & 0x0ff);
                RMAD2M = (break_points[2] >> 8) & 0x0ff;
                RMAD2H = (break_points[2] >> 16) & 0x0f;
                AIER20 = 1;
            }
            if (!(break_points[3] & 0x80000000))
            {
                RMAD3L = (break_points[3] & 0x0ff);
                RMAD3M = (break_points[3] >> 8) & 0x0ff;
                RMAD3H = (break_points[3] >> 16) & 0x0f;
                AIER21 = 1;
            }
#endif /* HAS_4BREAKPOINTS */
        }
        switch_to_user_mode();
    }
}

/**********************************************************************\
|*
|*  Function: m16c_stopped_at
|*
|*  Description: inform user where we stopped (current PC)
|*
 */
void m16c_stopped_at( void )
{
    uart_putchar('(');
    uart_putchar(ERR_PGM_STOPPED);
    m16c_dump_reg(0x11); /* show current break address to user */
    uart_putchar(')');
}
